# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-06-01.
#=============================================================================
#   tagexpression.py --- Tags and tag expressions
#=============================================================================
from __future__ import print_function, unicode_literals

import re

from six.moves import map

from .compat import unittest
from .filtersuite import flatten_suite
from .tags import get_tags


class TagExpressionParseError(Exception): pass

class TagExpression(object):
    tag_re = re.compile(r'@([\w_]+)$')
    negated_tag_re = re.compile(r'~@([\w_]+)$')
    
    def __init__(self, string=None):
        self.groups = None
        self.and_node = None
        if string:
            self.parse_and_add(string)
    
    def match(self, value):
        if self.and_node is None:
            return True
        else:
            return self.and_node.match(set(value))
    
    def filter_suite(self, suite, testSuiteClass=unittest.TestSuite):
        return testSuiteClass(
            test_item for test_item in flatten_suite(suite)
            if self.match(get_tags(test_item))
        )
    
    def parse_and_add(self, string):
        if self.and_node is None:
            self.and_node = AndNode()
        self.and_node.add_child(self.parse(string))
    
    def parse(self, string):
        tags = string.split(',')
        if len(tags) == 1:
            tag = tags[0]
            match = self.tag_re.match(tag.strip())
            if match:
                return TagNode(match.group(1))
            match = self.negated_tag_re.match(tag.strip())
            if match:
                return NegationNode(TagNode(match.group(1)))
            raise TagExpressionParseError(string)
        elif len(tags) == 0:
            raise TagExpressionParseError(string)
        else:
            return OrNode(list(map(self.parse, tags)))


class TagNode(object):
    
    def __init__(self, name):
        self.name = name
    
    def match(self, value):
        return self.name in value


class TrueNode(object):
    
    def match(self, value):
        return True


class FalseNode(object):
    
    def match(self, value):
        return False


class NegationNode(object):
    
    def __init__(self, child):
        self.child = child
    
    def match(self, value):
        return not self.child.match(value)


class AndNode(object):
    
    def __init__(self, child_nodes=[]):
        self.child_nodes = list(child_nodes)
    
    def add_child(self, child):
        self.child_nodes.append(child)
    
    def match(self, value):
        for child in self.child_nodes:
            if not child.match(value):
                return False
        return True


class OrNode(object):
    
    def __init__(self, child_nodes):
        self.child_nodes = tuple(child_nodes)
    
    def match(self, value):
        for child in self.child_nodes:
            if child.match(value):
                return True
        return False

#.............................................................................
#   tagexpression.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-06-01.
#=============================================================================
#   tags.py --- Add tags to tests
#=============================================================================
from __future__ import print_function, unicode_literals

import operator
from functools import reduce

from .compat import unittest


def tag(*tags):
    def decorator(test_item):
        test_item.__autocheck_tags__ = get_tags(test_item) | set(tags)
        return test_item
    return decorator

def add_tags(test_item, tags):
    all_tags = tuple(get_tags(test_item) | set(tags))
    return tag(*all_tags)(test_item)

def get_tags(test_item):
    tags = set()
    try:
        tags |= test_item.__autocheck_tags__
    except AttributeError:
        pass
    try:
        test_item = getattr(test_item, test_item._testMethodName)
    except AttributeError:
        pass
    try:
        test_item.__mro__
    except AttributeError:
        pass
    else:
        tags = reduce(operator.__or__,
            (getattr(cls, '__autocheck_tags__', set()) for cls in test_item.__mro__), tags)
    tags |= getattr(test_item, '__autocheck_tags__', set())
    return tags

#.............................................................................
#   tags.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-11.
#=============================================================================
#   tracing.py --- Debug tracing support
#=============================================================================
from __future__ import print_function, unicode_literals

import functools
import logging
import types

import six


trace = logging.getLogger(__name__)
trace.debug('enabled')


def tracing(f, name, log=trace.debug):
    @functools.wraps(f)
    def traced_f(*args, **kwargs):
        log('[%#x]%s(%s,%s)', id(args[0]), name, args[1:], kwargs)
        result = f(*args, **kwargs)
        log('[%#x]%s(%s,%s) => %r', id(args[0]), name, args[1:], kwargs, result)
        return result
    return traced_f

class MetaTracer(type):
    
    def __new__(self, classname, bases, classdict):
        log = logging.getLogger('%s.%s' % (classdict['__module__'], classname))
        classdict.setdefault('log', log)
        if trace.isEnabledFor(logging.DEBUG):
            for f in classdict:
                m = classdict[f]
                if isinstance(m, types.FunctionType):
                    classdict[f] = tracing(m, '%s.%s' % (classname, f), log=log.debug)
        return type.__new__(self, classname, bases, classdict)

class Tracer(six.with_metaclass(MetaTracer)):
    pass

def setLevel(level):
    trace.setLevel(level)

#.............................................................................
#   tracing.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-29.
#=============================================================================
#   status.py --- Test result status
#=============================================================================
from __future__ import print_function, unicode_literals


class Status(object):
    ordered = list()
    
    def __init__(self, key, name, name_plural, colour):
        self.key, self.name, self.name_plural, self.colour = key, name, name_plural, colour
        self.ordered.append(self)


ok = Status('.', 'ok', None, 'ok')
fail = Status('F', 'FAIL', 'failures', 'fail')
error = Status('E', 'ERROR', 'errors', 'error')
skip = Status('s', 'skipped', 'skipped', 'skip')
expected_failure = Status('x', 'expected failure', 'expectedFailures', 'expected_failure')
unexpected_success = Status('u', 'unexpected success', 'unexpectedSuccesses', 'unexpected_success')

#.............................................................................
#   status.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-28.
#=============================================================================
#   db.py --- Tests database
#=============================================================================
from __future__ import print_function, unicode_literals

import datetime
import functools
import hashlib
import inspect
import os
import re
import shutil
import sqlite3
from contextlib import contextmanager

import six

from .status import Status, ok


try:
    import pytz
except ImportError:
    have_pytz = False
else:
    have_pytz = True


class DatabaseError(Exception): pass
class DoesNotExist(DatabaseError): pass
class RunDoesNotExist(DoesNotExist): pass
class ResultDoesNotExist(DoesNotExist): pass
class TestDoesNotExist(DoesNotExist): pass

class Database(object):
    _TABLES = dict(
        
        run = '''run(
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            started TIMESTAMP,
            finished TIMESTAMP,
            wasSuccessful BOOLEAN,
            testsRun INTEGER,
            full BOOLEAN,
            errors INTEGER,
            failures INTEGER,
            skipped INTEGER,
            expectedFailures INTEGER,
            unexpectedSuccesses INTEGER
        )''',
        
        test = '''test(
            name VARCHAR PRIMARY KEY,
            test VARCHAR,
            suite VARCHAR,
            hash VARCHAR,
            runs INTEGER,
            average_time TIMEDELTA
        )''',
        
        result = '''result(
            name VARCHAR REFERENCES test(name),
            started TIMESTAMP,
            finished TIMESTAMP,
            run_id INTEGER REFERENCES run(id),
            status VARCHAR
        )''',
        
        version = '''version(
            id VARCHAR PRIMARY KEY
        )''',
    
    )
    name_re = re.compile(r'(?P<test>[^\s]+)\s+\((?P<suite>[^)]+)\)')
    version = '1'
    
    def __init__(self, path=None, basedir=None, name='.autocheck.db'):
        if path is None:
            basedir = os.getcwd() if basedir is None else basedir
            self.path = os.path.join(basedir, name)
        else:
            self.path = path
        self.connection = None
        self.current_run_id = None
    
    def _connect(self):
        self.connection = sqlite3.connect(self.path, detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES)
        self.connection.row_factory = sqlite3.Row
    
    def connect(self):
        self._connect()
        current_version = self.current_version()
        if current_version != self.version:
            self.migrate(current_version)
    
    def close(self):
        self.connection.close()
        self.connection = None
    
    @property
    def is_connected(self):
        return self.connection is not None
    
    def ensure_connection(self):
        if not self.is_connected:
            self.connect()
            self.setup()
    
    @contextmanager
    def transaction(self):
        self.ensure_connection()
        try:
            yield self.connection.cursor()
        except:
            self.connection.rollback()
            raise
        else:
            self.connection.commit()
    
    def with_cursor(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            cursor = kwargs.pop('cursor', None)
            if cursor is None:
                with self.transaction() as cursor:
                    return method(self, cursor, *args, **kwargs)
            else:
                return method(self, cursor, *args, **kwargs)
        return wrapper
    
    @with_cursor
    def current_version(self, cursor):
        self.create_table('version', cursor=cursor)
        try:
            cursor.execute('SELECT id FROM version ORDER BY id DESC LIMIT 1')
        except sqlite3.OperationalError:
            return '0'
        row = cursor.fetchone()
        if row is None:
            cursor.execute('INSERT INTO version(id) VALUES (?)', [self.version])
            return self.version
        return row[0]
    
    def migrate(self, current_version):
        print('migrating {} -> {} ...'.format(current_version, self.version), end='')
        self.close()
        new_path = '{}.bak-{}-{}'.format(self.path, current_version, self.version)
        os.rename(self.path, new_path)
        shutil.copy2(new_path, self.path)
        self._connect()
        if current_version == '0':
            self.migrate_0_1()
        print('done.')
    
    @with_cursor
    def migrate_0_1(self, cursor):
        self.create_table('version', cursor=cursor)
        cursor.execute('INSERT INTO version(id) VALUES (?)', [self.version])
        cursor.execute('ALTER TABLE test ADD COLUMN test VARCHAR')
        cursor.execute('ALTER TABLE test ADD COLUMN suite VARCHAR')
        cursor.execute('SELECT name FROM test')
        for row in cursor.fetchall():
            name = row[0]
            data = self.name_re.match(name).groupdict()
            data['name'] = name
            cursor.execute('UPDATE test SET test=:test, suite=:suite WHERE name=:name', data)
    
    @with_cursor
    def setup(self, cursor):
        cursor.execute('PRAGMA foreign_keys = ON')
        for name in self._TABLES:
            self.create_table(name, cursor=cursor)
    
    @with_cursor
    def create_table(self, cursor, name):
        cursor.execute('CREATE TABLE IF NOT EXISTS %s' % self._TABLES[name])
    
    @with_cursor
    def add_run(self, cursor):
        cursor.execute('INSERT INTO run(started) VALUES (?)', [self.utcnow()])
        self.current_run_id = cursor.lastrowid
    
    @property
    def total_runs(self):
        with self.transaction() as cursor:
            cursor.execute('SELECT count(*) FROM run')
            count = cursor.fetchone()[0]
        return count
    
    @with_cursor
    def get_run(self, cursor, run_id=None):
        if run_id is None:
            run_id = self.current_run_id
        if run_id is None:
            raise RunDoesNotExist()
        cursor.execute('SELECT * FROM run WHERE id = ?', [run_id])
        run = cursor.fetchone()
        if run is None:
            raise RunDoesNotExist(run_id)
        return run
    
    @with_cursor
    def get_test(self, cursor, name):
        cursor.execute('SELECT * FROM test WHERE name=?', (name,))
        test = cursor.fetchone()
        if test is None:
            raise TestDoesNotExist(name)
        return test
    
    @with_cursor
    def get_or_create_test(self, cursor, test_object):
        name = str(test_object)
        data = self.name_re.match(name).groupdict()
        try:
            test = self.get_test(name, cursor=cursor)
        except TestDoesNotExist:
            cursor.execute('INSERT INTO test(name,test,suite,hash,runs,average_time) VALUES (?,?,?,?,?,?)',
                (name,data['test'], data['suite'], source_hash(test_object), 0, 0))
            test = self.get_test(name, cursor=cursor)
        return test
    
    @with_cursor
    def add_result(self, cursor, test_object, started, finished, status):
        test = self.get_or_create_test(test_object, cursor=cursor)
        cursor.execute('INSERT INTO result(run_id,name,started,finished,status) VALUES (?,?,?,?,?)',
            (self.current_run_id, test[str('name')], started, finished, status))
        runs = test[str('runs')] + 1
        average_time = test[str('average_time')]
        if status == ok.key:
            if average_time is None:
                average_time = finished - started
            else:
                total_time = average_time * (runs - 1) + (finished - started)
                average_time = total_time / runs
        cursor.execute('UPDATE test SET runs=?,hash=?,average_time=? WHERE name=?',
            (runs, source_hash(test_object), average_time, str(test_object)))
    
    @with_cursor
    def add_results(self, cursor, results):
        for test_object, started, finished, status in results:
            self.add_result(test_object, self.to_utc(started), self.to_utc(finished), status, cursor=cursor)
    
    @with_cursor
    def get_last_result(self, cursor, name):
        cursor.execute('SELECT * FROM result WHERE name=? ORDER BY finished DESC LIMIT 1', (name,))
        result = cursor.fetchone()
        if result is None:
            raise ResultDoesNotExist(name)
        return result
    
    @with_cursor
    def total_runs_by_test_name(self, cursor, name):
        return self.get_test(name, cursor=cursor)[str('runs')]
    
    @with_cursor
    def get_result_count(self, cursor, run_id, status=None):
        if status is None:
            cursor.execute('SELECT count(*) FROM result WHERE run_id=?', [run_id])
        else:
            cursor.execute('SELECT count(*) FROM result WHERE status=? AND run_id=?', (status, run_id))
        return cursor.fetchone()[0]
    
    @with_cursor
    def get_result_counts(self, cursor, run_id):
        for status in Status.ordered:
            if status.name != ok.name:
                yield status.name_plural, self.get_result_count(run_id, status.key, cursor=cursor)
    
    @with_cursor
    def finish_run(self, cursor, full):
        run_id = self.current_run_id
        data = dict(self.get_result_counts(run_id, cursor=cursor))
        data.update(
            run_id = run_id,
            finished = self.utcnow(),
            wasSuccessful = data['errors'] == data['failures'] == 0,
            testsRun = self.get_result_count(run_id, cursor=cursor),
            full = full,
        )
        cursor.execute('''UPDATE run SET
            finished=:finished,
            wasSuccessful=:wasSuccessful,
            testsRun=:testsRun,
            full=:full,
            errors=:errors,
            failures=:failures,
            skipped=:skipped,
            expectedFailures=:expectedFailures,
            unexpectedSuccesses=:unexpectedSuccesses WHERE id=:run_id''', data)
        run = self.get_run(run_id, cursor=cursor)
        self.clean_history(cursor=cursor)
        return run
    
    @with_cursor
    def get_last_run_id(self, cursor, where=''):
        cursor.execute('SELECT id FROM run %s ORDER BY finished DESC LIMIT 1' % where)
        run_id = cursor.fetchone()
        if run_id is not None:
            return run_id[0]
    
    @with_cursor
    def get_last_successful_run_id(self, cursor):
        return self.get_last_run_id('WHERE wasSuccessful=1', cursor=cursor)
    
    @with_cursor
    def get_last_successful_full_run_id(self, cursor):
        return self.get_last_run_id('WHERE wasSuccessful=1 AND full=1', cursor=cursor)
    
    @with_cursor
    def get_last_run_ids(self, cursor):
        return (
            self.get_last_run_id(cursor=cursor),
            self.get_last_successful_run_id(cursor=cursor),
            self.get_last_successful_full_run_id(cursor=cursor),
        )
    
    @with_cursor
    def collect_results_after(self, cursor, run_id, status=ok.key, exclude=True):
        cursor.execute('''SELECT DISTINCT name FROM result WHERE run_id IN (
            SELECT id FROM run WHERE started>(
                SELECT started FROM run WHERE id=?
            )) AND status%s?''' % ('=', '!=')[bool(exclude)], (run_id, status))
        for row in cursor.fetchall():
            yield row[0]
    
    @with_cursor
    def failures(self, cursor):
        last, successful, full = self.get_last_run_ids(cursor=cursor)
        if last != successful:
            return set(self.collect_results_after(full, cursor=cursor))
        else:
            return set()
    
    @with_cursor
    def source_has_changed(self, cursor, test_object):
        new_hash = source_hash(test_object)
        try:
            return self.get_test(str(test_object), cursor=cursor)[str('hash')] != new_hash
        except TestDoesNotExist:
            return True
    
    @with_cursor
    def collect_changes(self, cursor, tests):
        for test_object in tests:
            if self.source_has_changed(test_object, cursor=cursor):
                yield str(test_object)
    
    @with_cursor
    def candidates(self, cursor, tests):
        return self.failures(cursor=cursor) | set(self.collect_changes(tests, cursor=cursor))
    
    @with_cursor
    def should_run_again(self, cursor):
        last, successful, full = self.get_last_run_ids(cursor=cursor)
        if last is None:
            return False
        elif last != successful:
            return False
        elif last == full:
            return False
        else:
            return True
    
    @with_cursor
    def clean_history(self, cursor):
        run_id = self.get_last_successful_full_run_id(cursor=cursor)
        cursor.execute('DELETE FROM result WHERE finished<(SELECT started FROM run WHERE id=?)', (run_id,))
        cursor.execute('DELETE FROM run WHERE (SELECT count(*) FROM result WHERE run_id=run.id)=0')
    
    @with_cursor
    def stats(self, cursor, suite=None):
        if suite is None:
            cursor.execute('SELECT * FROM test ORDER BY average_time')
        else:
            cursor.execute('SELECT * FROM test WHERE suite=? ORDER BY average_time', [suite])
        for row in cursor.fetchall():
            data = dict(zip(row.keys(), row))
            data['time'] = data['average_time'].total_seconds()
            yield data

    @with_cursor
    def stats_grouped(self, cursor):
        cursor.execute('SELECT suite, total(average_time) as time FROM test GROUP BY suite ORDER BY time')
        for row in cursor.fetchall():
            suite = dict(zip(row.keys(), row))
            suite['tests'] = list(self.stats(cursor=cursor, suite=suite['suite']))
            yield suite
    
    def to_utc(self, datetime):
        return pytz.utc.localize(datetime) if have_pytz else datetime
    
    def utcnow(self):
        return self.to_utc(datetime.datetime.utcnow())

def source_hash(test_object):
    test_method = getattr(test_object, test_object._testMethodName)
    try:
        source = test_method.im_self.getsource()
    except AttributeError:
        source = inspect.getsource(test_method)
    if six.PY3:
        source = source.encode('utf-8')
    return hashlib.sha1(source).hexdigest()

def timedelta_to_float(delta):
    if hasattr(delta, 'total_seconds'):
        return delta.total_seconds()
    else:
        return (delta.microseconds + (delta.seconds + delta.days * 24. * 3600.) * 10**6) / 10**6

def adapt_timedelta(delta):
    return str(timedelta_to_float(delta))

def convert_timedelta(s):
    return datetime.timedelta(seconds=float(s))

sqlite3.register_adapter(datetime.timedelta, adapt_timedelta)
sqlite3.register_converter(str('timedelta'), convert_timedelta)

def convert_boolean(s):
    return bool(int(s))

sqlite3.register_converter(str('boolean'), convert_boolean)

#.............................................................................
#   db.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-12-22.
#=============================================================================
#   gitignore.py --- Test files for if git ignores them
#=============================================================================
from __future__ import print_function, unicode_literals

import glob
import os

from .logger import Logger


class GitIgnore(Logger):
    
    def __init__(self, root):
        self.root = os.path.abspath(root)
        self._matchers = dict()

    def _match(self, dir, path):
        try:
            matchers = self._matchers[dir]
        except KeyError:
            matchers = self._matchers[dir] = self._make_matchers(dir)
        for matcher in matchers:
            if matcher(path):
                self.log.debug('_match:[%s][%s] => True', dir, path)
                return True
        self.log.debug('_match:[%s][%s] => False', dir, path)
        return False
    
    def _make_matchers(self, dir, gitignore=None):
        matchers = tuple()
        path = os.path.abspath(dir)
        if gitignore is None:
            gitignore = os.path.join(path, '.gitignore')
        for ignore in (gitignore, os.path.join(os.path.expanduser('~'), '.gitignore')):
            if os.path.exists(ignore):
                matchers += tuple(self._make_matchers_from_file(path, ignore))
        if path == self.root:
            exclude = os.path.join(path, '.git', 'info', 'exclude')
            if os.path.exists(exclude):
                matchers += tuple(self._make_matchers_from_file(path, exclude))
        return matchers
    
    def _make_matchers_from_file(self, dir, gitignore):
        for line in open(gitignore):
            line = line.strip()
            if line.startswith('#'):
                continue
            if not line:
                continue
            self.log.debug('%s: %s', gitignore, line)
            yield self._make_matcher_from_line(dir, line)
    
    def _make_matcher_from_line(self, dir, line):
        if line.endswith('/'):
            line = line.rstrip('/')
            isdir = True
        else:
            isdir = False
        if line.startswith('/'):
            line = line.lstrip('/')
            def makepattern(fullpath):
                return os.path.join(dir, line)
        else:
            def makepattern(fullpath):
                return os.path.join(os.path.dirname(fullpath), line)
        def matcher(path):
            fullpath = os.path.join(dir, path)
            self.log.debug('fullpath: %r', fullpath)
            pattern = makepattern(fullpath)
            self.log.debug('globbing: %r', pattern)
            globs = glob.glob(pattern)
            self.log.debug('globs: %s', globs)
            if isdir:
                for ignoredir in globs:
                    if not os.path.isdir(ignoredir):
                        continue
                    self.log.debug('%s %s %s', os.path.commonprefix((fullpath, ignoredir)), fullpath, ignoredir)
                    if os.path.commonprefix((fullpath, ignoredir)) == ignoredir:
                        return True
                return False
            return fullpath in globs
        return matcher
    
    def match(self, path):
        dir = os.path.abspath(os.path.dirname(path))
        path = os.path.basename(path)
        while dir != self.root:
            if self._match(dir, path):
                return True
            path = os.path.join(os.path.basename(dir), path)
            dir = os.path.dirname(dir)
        if self._match(self.root, path):
            return True
        return False

#.............................................................................
#   gitignore.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-12.
#=============================================================================
#   testrunner.py --- Running tests
#=============================================================================
from __future__ import print_function, unicode_literals

import datetime
import logging
import sys
import time
from io import BytesIO

import six
from six.moves import map

from . import status
from .colorizer import ColourWritelnDecorator, ColourScheme
from .compat import unittest
from .filtersuite import filter_suite
from .tagexpression import TagExpression


try:
    from .growler import Notifier
except ImportError:
    growler = None
else:
    growler = Notifier()


def camel_to_underscore(value):
    def camel_to_underscore():
        yield value[0]
        for c, d in zip(value[1:], value[2:]+'_'):
            if c.isupper() and not (d == '_' or d.isupper()):
                yield '_'
                yield c.lower()
            else:
                yield c
    return ''.join(camel_to_underscore())

class LogHandler(logging.StreamHandler):
    
    def __init__(self):
        logging.Handler.__init__(self)
    
    @property
    def stream(self):
        return sys.stderr

class TestResult(unittest.TestResult):
    """A test result class that can print formatted text results to a stream.
    
    Used by TestRunner.
    """
    separator1 = '=' * 70
    separator2 = '-' * 70
    
    def __init__(self, stream=sys.stderr, descriptions=True, verbosity=1,
            colourscheme='light', database=None, now_function=None):
        super(TestResult, self).__init__()
        self.scheme = colourscheme if isinstance(colourscheme, ColourScheme) else ColourScheme(colourscheme)
        self.stream = stream if isinstance(stream, ColourWritelnDecorator) else ColourWritelnDecorator(stream)
        self.showAll = verbosity > 1
        self.specs = descriptions and self.showAll
        self.dots = verbosity == 1
        self.descriptions = descriptions
        self.success_count = 0
        self.current_class = None
        self.database = database
        self.now = datetime.datetime.utcnow if now_function is None else now_function
    
    def getSpecDescription(self, test):
        if self.specs:
            doc_first_line = test.shortDescription()
            if self.descriptions and doc_first_line:
                spec = doc_first_line
            else:
                spec = self._specify(test)
            if test.__class__ != self.current_class:
                self.current_class = test.__class__
                prefix = self._specify_class()
            else:
                prefix = '  '
            return prefix + spec
        else:
            return self.getDescription(test)
    
    def getDescription(self, test):
        doc_first_line = test.shortDescription()
        if self.descriptions and doc_first_line:
            return '\n'.join((str(test), doc_first_line))
        else:
            return str(test)
    
    def startTestRun(self):
        super(TestResult, self).startTestRun()
        self.results = []
    
    def stopTestRun(self):
        super(TestResult, self).stopTestRun()
        if self.database is not None and self.results:
            self.database.add_results(self.results)
    
    def startTest(self, test):
        super(TestResult, self).startTest(test)
        if self.showAll:
            self.stream.write(self.getSpecDescription(test))
            self.stream.write(' ... ')
            if self.specs and getattr(test, 'is_scenario', False):
                self.stream.writeln()
            self.stream.flush()
        self.time_started = self.now()
    
    def addSuccess(self, test):
        super(TestResult, self).addSuccess(test)
        self._add_result(test, status.ok.key)
        self.success_count += 1
        self._write_scenario_result_indent(test)
        self._write_status(status.ok)
    
    def addError(self, test, err):
        super(TestResult, self).addError(test, err)
        self._add_result(test, status.error.key)
        self._write_scenario_result_indent(test)
        self._write_status(status.error)
    
    def addFailure(self, test, err):
        super(TestResult, self).addFailure(test, err)
        self._add_result(test, status.fail.key)
        self._write_scenario_result_indent(test)
        self._write_status(status.fail)
    
    def addSkip(self, test, reason):
        super(TestResult, self).addSkip(test, reason)
        self._add_result(test, status.skip.key)
        colour = self.scheme.skip
        if self.showAll:
            self._write_scenario_result_indent(test)
            self._write('{0} {1!r}'.format(status.skip.name, reason), None, colour=colour)
        elif self.dots:
            self._write(None, status.skip.key, colour=colour)
    
    def addExpectedFailure(self, test, err):
        super(TestResult, self).addExpectedFailure(test, err)
        self._add_result(test, status.expected_failure.key)
        if self.showAll:
            self._write_scenario_result_indent(test)
        self._write_status(status.expected_failure)
    
    def addUnexpectedSuccess(self, test):
        super(TestResult, self).addUnexpectedSuccess(test)
        self._add_result(test, status.unexpected_success.key)
        if self.showAll:
            self._write_scenario_result_indent(test)
        self._write_status(status.unexpected_success)
    
    def startStep(self, step):
        if self.specs and self.showAll:
            self.stream.write('    %s ... ' % step)
            self.stream.flush()
    
    def stopStep(self, step):
        if self.specs and self.showAll:
            self.stream.writeln()
    
    def addStepSuccess(self, step):
        if self.specs and self.showAll:
            self.stream.write('ok', colour=self.scheme.ok)
    
    def addStepError(self, step):
        if self.specs and self.showAll:
            self.stream.write('error', colour=self.scheme.error)
    
    def addStepFailure(self, step):
        if self.specs and self.showAll:
            self.stream.write('fail', colour=self.scheme.fail)
    
    def addStepUndefined(self, step):
        if self.specs and self.showAll:
            self.stream.write('undefined', colour=self.scheme.skip)
    
    def printErrors(self):
        if self.dots or self.showAll:
            self.stream.writeln()
        self.printErrorList('ERROR', self.errors, colour=self.scheme.error)
        self.printErrorList('FAIL', self.failures, colour=self.scheme.fail)
    
    def printErrorList(self, flavour, errors, colour):
        for test, err in errors:
            self.stream.writeln(self.separator1)
            self.stream.writeln('%s: %s' % (flavour, self.getDescription(test)), colour=colour)
            self.stream.writeln(self.separator2)
            self.stream.writeln('%s' % err, colour=colour)
    
    
    def _add_result(self, test, status):
        if self.database is not None:
            self.results.append((test, self.time_started, self.now(), status))
    
    def _specify(self, test):
        if test._testMethodName.startswith('test_'):
            return test._testMethodName[5:].replace('_', ' ')
        else:
            return str(test)
    
    def _specify_class(self):
        name = self.current_class.__name__
        if name.endswith('Test') or name.endswith('Vows'):
            name = name[:-4]
        return '\n%s:\n  ' % ' '.join(camel_to_underscore(name).replace('_', ' ').split())
    
    def _write(self, all, dots, colour):
        if self.showAll and all is not None:
            self.stream.writeln(all, colour=colour)
        elif self.dots and dots is not None:
            self.stream.write(dots, colour=colour)
            self.stream.flush()
    
    def _write_status(self, status):
        self._write(status.name, status.key, getattr(self.scheme, status.colour))
    
    def _write_scenario_result_indent(self, test):
        if self.specs and getattr(test, 'is_scenario', False):
            self.stream.write('  ... ')
    
    def _setupStdout(self):
        if self.buffer:
            if self._stderr_buffer is None:
                self._stderr_buffer = UTF8StringIO()
                self._stdout_buffer = UTF8StringIO()
            sys.stdout = self._stdout_buffer
            sys.stderr = self._stderr_buffer


class UTF8StringIO(BytesIO):
    
    def write(self, s):
        if not isinstance(s, six.string_types):
            s = six.text_type(s)
        if isinstance(s, six.text_type):
            s = s.encode('utf-8')
        super(UTF8StringIO, self).write(s)


class TestRunner(object):
    """A test runner class that displays results in textual form.
    
    It prints out the names of tests as they are run, errors as they
    occur, and a summary of the results at the end of the test run.
    """
    resultclass = TestResult
    database = None
    
    def __init__(self, stream=sys.stderr, descriptions=True, verbosity=1,
                 failfast=False, buffer=False, resultclass=None,
                 colourscheme='light', growler=growler, database=None,
                 full_suite=True, now_function=None):
        self.stream = ColourWritelnDecorator(stream)
        self.scheme = ColourScheme(colourscheme)
        self.descriptions = descriptions
        self.verbosity = verbosity
        self.failfast = failfast
        self.buffer = buffer
        self.growler = growler
        if resultclass is not None:
            self.resultclass = resultclass
        if database is not None:
            self.database = database
        self.full_suite = full_suite
        self.now = datetime.datetime.utcnow if now_function is None else now_function
    
    def _makeResult(self):
        return self.resultclass(self.stream, self.descriptions, self.verbosity, self.scheme, self.database, self.now)
    
    def run(self, test):
        "Run the given test case or test suite."
        result = self._makeResult()
        unittest.registerResult(result)
        result.failfast = self.failfast
        result.buffer = self.buffer
        if self.database is not None:
            self.database.add_run()
        startTime = time.time()
        startTestRun = getattr(result, 'startTestRun', None)
        if startTestRun is not None:
            startTestRun()
        try:
            test(result)
        finally:
            stopTestRun = getattr(result, 'stopTestRun', None)
            if stopTestRun is not None:
                stopTestRun()
        stopTime = time.time()
        if self.database is not None:
            self.database.finish_run(self.full_suite)
        timeTaken = stopTime - startTime
        result.printErrors()
        if hasattr(result, 'separator2'):
            self.stream.writeln(result.separator2)
        run = result.testsRun
        self.stream.writeln('Ran %d test%s in %.3fs' %
                            (run, run != 1 and 's' or '', timeTaken))
        self.stream.writeln()
        
        expectedFails = unexpectedSuccesses = skipped = 0
        try:
            results = map(len, (result.expectedFailures,
                                result.unexpectedSuccesses,
                                result.skipped))
        except AttributeError:
            pass
        else:
            expectedFails, unexpectedSuccesses, skipped = results
        
        infos = []
        if not result.wasSuccessful():
            self.stream.write('FAILED', colour=self.scheme.FAIL)
            failed, errored = map(len, (result.failures, result.errors))
            if failed:
                infos.append((self.scheme.FAIL, 'failures: %d' % failed))
            if errored:
                infos.append((self.scheme.ERROR, 'errors: %d' % errored))
        else:
            failed = errored = None
            self.stream.write('OK', colour=self.scheme.OK)
        if skipped:
            infos.append((self.scheme.SKIP, 'skipped: %d' % skipped))
        if expectedFails:
            infos.append((
                self.scheme.EXPECTED_FAILURE,
                'expected failures: %d' % expectedFails))
        if unexpectedSuccesses:
            infos.append((
                self.scheme.UNEXPECTED_SUCCESS,
                'unexpected successes: %d' % unexpectedSuccesses))
        try:
            success = result.success_count
        except AttributeError:
            pass
        else:
            if not result.wasSuccessful():
                infos.append((self.scheme.OK, 'passed: %d' % success))
        if self.growler is not None:
            if errored:
                kind = 'error'
            elif failed:
                kind = 'fail'
            elif skipped or expectedFails:
                kind = 'pending'
            else:
                assert result.wasSuccessful()
                kind = 'pass'
            description = ['%d tests run' % run]
            if infos:
                description.append('(%s)' % ', '.join(info for _, info in infos))
            self.growler.notify(kind.title(), '\n'.join(description), dict(error='fail').get(kind, kind))
        if infos:
            self.stream.writeln(' (%s)' % (', '.join(c(info) for c, info in infos),))
        else:
            self.stream.writeln()
        return result


class TestProgram(unittest.TestProgram):
    
    def __init__(self, module='__main__', defaultTest=None, argv=None,
                    testRunner=None, testLoader=unittest.defaultTestLoader,
                    exit=True, verbosity=1, failfast=None, catchbreak=None,
                    buffer=None, database=None):
        self.database = database
        self.run_tags = None
        super(TestProgram, self).__init__(module=module, defaultTest=defaultTest,
            argv=argv, testRunner=testRunner, testLoader=testLoader, exit=exit,
            verbosity=verbosity, failfast=failfast, catchbreak=catchbreak,
            buffer=buffer)
    
    def parseArgs(self, argv):
        new_args = []
        i = 0
        while i < len(argv):
            arg = argv[i]
            if arg.startswith('--tags'):
                if self.run_tags is None:
                    self.run_tags = TagExpression()
                if arg.startswith('--tags='):
                    arg = arg[7:]
                else:
                    i += 1
                    arg = argv[i]
                self.run_tags.parse_and_add(arg)
            else:
                new_args.append(arg)
            i += 1
        super(TestProgram, self).parseArgs(new_args)
    
    def runTests(self):
        if self.catchbreak:
            unittest.installHandler()
        if self.run_tags is not None:
            self.test = self.run_tags.filter_suite(self.test)
        tests_to_run, full_suite = filter_suite(self.test, self.database)
        testRunner = self.testRunner(verbosity=self.verbosity,
            failfast=self.failfast, buffer=self.buffer,
            database=self.database, full_suite=full_suite)
        self.result = testRunner.run(tests_to_run)
        if self.exit:
            sys.exit(not self.result.wasSuccessful())


if __name__ == '__main__':
    import sys
    if sys.argv[0].endswith('__main__.py'):
        sys.argv[0] = 'python -m autocheck.testrunner'
    
    from unittest.main import USAGE_AS_MAIN
    TestProgram.USAGE = USAGE_AS_MAIN
    
    main(module=None, testRunner=TestRunner)


__test__ = False

#.............................................................................
#   testrunner.py

# Derived from Python-2.7.1 unittest.runner, unittest.result, traceback
# Copyright © 2001-2010 Python Software Foundation; All Rights Reserved
# http://docs.python.org/license.html



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2015-04-19.
#=============================================================================
#   quickstart.py --- Quick test database setup
#=============================================================================
"""
Extremely fast Django test runner, based on the idea that your database schema
and fixtures are changed much more seldom that your code and tests.  All you
need is to make sure that your "quickstart.sqlite" database file is always up
to date.

BEWARE: Don't run this test runner on production server. It assumes that you 
use only one database configured as "default", and its db engine is SQLite.
Otherwise your tests can eat your data!

How to use it:

1. Install sqlitebck:

    pip install sqlitebck

2. Create database alias named "quickstart" with sqlite3 backend in your
   settings.py:

    DATABASES = {
         ..
        'quickstart': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': '.../quickstart.sqlite',
        },
    }

3. Create quickstart.sqlite:
     
     ./manage.py migrate --database quickstart

4. Redefine variable TEST_RUNNER in the settings.py:

    TEST_RUNNER = 'autocheck.contrib.django.TestSuiteRunner'

5. If your schema and has changed, apply migrations to quickstart.sqlite:

     ./manage.py migrate --database quickstart

Original code and docstring: https://gist.github.com/NetAngels/1044215

We use sqlitebck: https://pypi.python.org/pypi/sqlitebck instead of apsw.
"""
from django.db import connections


try:
    import sqlitebck
except ImportError:
    sqlitebck = None


class QuickstartMixin(object):

    def setup_databases(self, **kwargs):
        if sqlitebck is not None and 'quickstart' in connections:
            quickstart = connections['quickstart']
            connection = connections['default']
            quickstart.ensure_connection()
            connection.ensure_connection()
            sqlitebck.copy(quickstart.connection, connection.connection)
            serialize = connection.settings_dict.get("TEST", {}).get("SERIALIZE", True)
            if serialize:
                connection._test_serialized_contents = connection.creation.serialize_db_to_string()
            return 'quickstart'
        else:
            return super(QuickstartMixin, self).setup_databases(**kwargs)

    def teardown_databases(self, old_config, **kwargs):
        if old_config != 'quickstart':
            super(QuickstartMixin, self).teardown_databases(old_config, **kwargs)

#.............................................................................
#   quickstart.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-12.
#=============================================================================
#   test_colorizer.py --- Colorizer vows
#=============================================================================
from __future__ import unicode_literals

from io import StringIO

import termstyle
from should_dsl import should

from autocheck.colorizer import ColourWritelnDecorator, ColourDecorator, ColourScheme
from autocheck.compat import unittest
from autocheck.testrunner import TestResult


class ColourTestCase(unittest.TestCase):
    
    @classmethod
    def setUpClass(self):
        termstyle.Style.enabled |should| be(True)

class ColourDecoratorTest(ColourTestCase):
    
    def test_writes_arg_unchanged_without_colour(self):
        stream = ColourDecorator(StringIO())
        stream.write('{*}')
        stream.stream.getvalue() |should| be_equal_to('{*}')
    
    def test_wraps_arg_in_colour(self):
        stream = ColourDecorator(StringIO())
        stream.write('{*}', colour=termstyle.italic)
        stream.stream.getvalue() |should| be_equal_to(termstyle.italic('{*}'))


class ColourWritelnDecoratorTest(ColourTestCase):
    
    def test_writes_arg_as_line_without_colour(self):
        stream = ColourWritelnDecorator(StringIO())
        stream.writeln('{*}')
        stream.stream.getvalue() |should| be_equal_to('{*}\n')
    
    def test_wraps_arg_as_line_in_colour(self):
        stream = ColourWritelnDecorator(StringIO())
        stream.writeln('{*}', colour=termstyle.red)
        stream.stream.getvalue() |should| be_equal_to(termstyle.red('{*}') + '\n')


class ColourSchemeTest(ColourTestCase):
    
    def test_translates_names_to_termstyle_colours(self):
        scheme = ColourScheme(dict(ok='green'))
        scheme.ok |should| be(termstyle.green)
    
    def test_translates_sequences_of_styles_single_callable(self):
        scheme = ColourScheme(dict(blue_italic=('blue', 'italic')))
        scheme.blue_italic('{*}') |should| be_equal_to(termstyle.italic(termstyle.blue('{*}')))
    
    def test_can_load_colours_from_yaml_file_by_scheme_name(self):
        scheme = ColourScheme('light')
        scheme.ok |should| be(termstyle.green)
    
    def test_returns_always_None_if_no_scheme_given(self):
        scheme = ColourScheme(None)
        scheme.ok |should| be(None)



class ColourTestResultTest(ColourTestCase):
    
    def test_is_subclass_of_unittest_TestResult(self):
        TestResult() |should| be_instance_of(unittest.TestResult)
    
    def test_accepts_colourscheme_keyword(self):
        result = TestResult(colourscheme=None)
    
    def test_uses_colourscheme_from_yaml_file(self):
        result = TestResult(colourscheme='light')
        result.scheme.ok |should| be_equal_to(termstyle.green)


#.............................................................................
#   test_colorizer.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-06-01.
#=============================================================================
#   test_tagexpression.py --- Tag expression vows
#=============================================================================
import os.path

import six
import yaml
from should_dsl import should

from autocheck.compat import unittest
from autocheck.tagexpression import TagExpression, TagNode, TrueNode, FalseNode, NegationNode, AndNode, OrNode


with open(os.path.splitext(__file__)[0] + '.yaml') as f:
    test_examples = yaml.load(f)


class TagExpressionVowsMeta(type):
    
    def __new__(self, classname, bases, classdict):
        for key, test in six.iteritems(test_examples):
            for test_in, test_result in test['examples']:
                test_name = '_'.join(['test', 'example', 'with', key.replace(' ', '_')]
                    + (test_in if test_in else ['empty']))
                define_method(classdict, test['groups'], test_in, test_result, test_name)
        return type.__new__(self, classname, bases, classdict)

def define_method(classdict, groups, test_in, test_result, test_name):
    def runTest(self):
        tags = TagExpression()
        for group in groups:
            tags.parse_and_add(group)
        tags.match(test_in) |should| be(test_result)
    runTest.__name__ = test_name
    classdict[test_name] = runTest


class TagExpressionVows(six.with_metaclass(TagExpressionVowsMeta, unittest.TestCase)):
    
    def test_always_matches_when_empty(self):
        tags = TagExpression()
        result = tags.match(())
        result |should| be(True)
    
    def test_parses_tag(self):
        tags = TagExpression()
        node = tags.parse('@tag_name')
        node |should| be_instance_of(TagNode)
        node.name |should| be_equal_to('tag_name')
    
    def test_parses_negated_tag(self):
        tags = TagExpression()
        node = tags.parse('~@tag_name')
        node |should| be_instance_of(NegationNode)
        node.child |should| be_instance_of(TagNode)
        node.child.name |should| be_equal_to('tag_name')
    
    def test_parses_comma_separated_group_as_or_expression(self):
        tags = TagExpression()
        node = tags.parse('@tag_1,~@tag_2,@tag_3')
        
        node |should| be_instance_of(OrNode)
        
        node_1, node_2, node_3 = node.child_nodes
        
        node_1 |should| be_instance_of(TagNode)
        node_1.name |should| be_equal_to('tag_1')
        
        node_2 |should| be_instance_of(NegationNode)
        node_2.child |should| be_instance_of(TagNode)
        node_2.child.name |should| be_equal_to('tag_2')
        
        node_3 |should| be_instance_of(TagNode)
        node_3.name |should| be_equal_to('tag_3')
    
    def test_combines_parsed_expressions_to_and_expression(self):
        tags = TagExpression()
        tags.parse_and_add('@tag_0')
        
        tags.and_node |should| be_instance_of(AndNode)
        child_nodes = tags.and_node.child_nodes
        len(child_nodes) |should| be(1)
        child_nodes[0] |should| be_instance_of(TagNode)
        child_nodes[0].name |should| be_equal_to('tag_0')


class TagNodeVows(unittest.TestCase):
    
    def test_matches_when_value_is_in_testee(self):
        node = TagNode('name')
        result = node.match(('name', 'other'))
        result |should| be(True)
    
    def test_does_not_matche_when_value_is_not_in_testee(self):
        node = TagNode('name')
        result = node.match(('other', 'value'))
        result |should| be(False)


class TrueNodeVows(unittest.TestCase):
    
    def test_always_matches(self):
        node = TrueNode()
        result = node.match(())
        result |should| be(True)


class FalseNodeVows(unittest.TestCase):
    
    def test_never_matches(self):
        node = FalseNode()
        result = node.match(())
        result |should| be(False)


class NegationNodeVows(unittest.TestCase):
    
    def test_matches_when_child_does_not_match(self):
        node = NegationNode(FalseNode())
        result = node.match(())
        result |should| be(True)
    
    def test_does_not_match_when_child_does(self):
        node = NegationNode(TrueNode())
        result = node.match(())
        result |should| be(False)


class AndNodeVows(unittest.TestCase):
    
    def test_matches_when_all_child_nodes_do(self):
        node = AndNode((TrueNode(), TrueNode()))
        result = node.match(())
        result |should| be(True)
    
    def test_does_not_match_when_one_child_does_not(self):
        node = AndNode((TrueNode(), FalseNode(), TrueNode()))
        result = node.match(())
        result |should| be(False)
    
    def test_can_add_child_node(self):
        node = AndNode()
        node_1, node_2 = TrueNode(), FalseNode()
        node.add_child(node_1)
        node.add_child(node_2)
        node.child_nodes |should| be_equal_to([node_1, node_2])


class OrNodeVows(unittest.TestCase):
    
    def test_matches_when_at_least_one_child_does(self):
        node = OrNode((FalseNode(), TrueNode(), FalseNode()))
        result = node.match(())
        result |should| be(True)
    
    def test_does_not_match_when_no_child_does(self):
        node = AndNode((FalseNode(), FalseNode()))
        result = node.match(())
        result |should| be(False)


#.............................................................................
#   test_tagexpression.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-29.
#=============================================================================
#   filtersuite.py --- Filter test suites
#=============================================================================
from __future__ import print_function, unicode_literals

from operator import methodcaller

from .compat import unittest


def flatten_suite(suite):
    def flatten(suite):
        for test in suite:
            if isinstance(test, unittest.TestSuite):
                for t in flatten(test):
                    yield t
            else:
                yield test
    return unittest.TestSuite(sorted(flatten(suite), key=methodcaller('id')))


def filter_suite(suite, database):
    full_suite = True
    suite = flatten_suite(suite)
    if database is not None:
        candidates = database.candidates(suite)
        if candidates:
            tests_to_run = [test for test in suite if str(test) in candidates]
            if tests_to_run:
                suite = unittest.TestSuite(tests_to_run)
                full_suite = False
    return suite, full_suite


#.............................................................................
#   filtersuite.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2015-04-19.
#=============================================================================
#   __init__.py --- Django support
#=============================================================================
import os.path


try:
    from .discoveryrunner import TestSuiteRunner
except ImportError:
    try:
        from .testsuiterunner import TestSuiteRunner
    except ImportError:
        TestSuiteRunner = None

if TestSuiteRunner is None:
    def is_django():
        return False
else:
    try:
        from south.management.commands import patch_for_test_db_setup
    except ImportError:
        pass
    else:
        patch_for_test_db_setup()

    def is_django():
        if os.path.exists('manage.py'):
            with open('manage.py') as manage:
                for line in manage:
                    if 'DJANGO_SETTINGS_MODULE' in line:
                        return True

#.............................................................................
#   __init__.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-29.
#=============================================================================
#   test_testrunner.py --- TestResult vows
#=============================================================================
from __future__ import unicode_literals

import datetime
import sys
from functools import partial
from io import StringIO

import mock
from should_dsl import should_not

from autocheck.compat import unittest
from autocheck.status import expected_failure, ok, error, fail, skip, unexpected_success
from autocheck.testrunner import TestResult, TestRunner


class TestTestCase(unittest.TestCase):
    
    class Test(unittest.TestCase):
        def runTest(self):
            pass
    
    def setUp(self):
        self.test = self.Test()
        self.stream = StringIO()


@mock.patch('autocheck.db.Database')
class DatabaseTestResultVows(TestTestCase):
    
    def test_accepts_database_parameter(self, Database):
        db = Database()
        constructor = partial(TestResult, database=db, stream=self.stream)
        
        constructor |should_not| throw(TypeError)
    
    def test_accepts_now_function_parameter(self, Database):
        def now_function():
            pass
        constructor = partial(TestResult, now_function=now_function, stream=self.stream)
        
        constructor |should_not| throw(TypeError)
    
    def start_test(self, db):
        start = datetime.datetime.utcnow()
        finish = start + datetime.timedelta(seconds=.1)
        now = mock.Mock('now', return_value=start)
        result = TestResult(database=db, stream=self.stream, now_function=now)
        result.startTestRun()
        result.startTest(self.test)
        now.return_value = finish
        return result, start, finish
    
    def stop_test(self, result):
        result.stopTest(self.test)
        result.stopTestRun()
    
    def test_passes_on_success_to_database(self, Database):
        db = Database()
        result, start, finish = self.start_test(db)
        
        result.addSuccess(self.test)
        self.stop_test(result)
        
        db.add_results.assert_called_once_with([(self.test, start, finish, ok.key)])
    
    def test_passes_on_failure_to_database(self, Database):
        db = Database()
        result, start, finish = self.start_test(db)

        try: x
        except Exception as e:
            result.addFailure(self.test, sys.exc_info())
        self.stop_test(result)

        db.add_results.assert_called_once_with([(self.test, start, finish, fail.key)])

    def test_passes_on_error_to_database(self, Database):
        db = Database()
        result, start, finish = self.start_test(db)

        try: x
        except Exception as e:
            result.addError(self.test, sys.exc_info())
        self.stop_test(result)

        db.add_results.assert_called_once_with([(self.test, start, finish, error.key)])

    def test_passes_on_expected_failure_to_database(self, Database):
        db = Database()
        result, start, finish = self.start_test(db)

        try: x
        except Exception as e:
            result.addExpectedFailure(self.test, sys.exc_info())
        self.stop_test(result)

        db.add_results.assert_called_once_with([(self.test, start, finish, expected_failure.key)])
    
    def test_passes_on_skip_to_database(self, Database):
        db = Database()
        result, start, finish = self.start_test(db)
        
        result.addSkip(self.test, 'reason')
        self.stop_test(result)
        
        db.add_results.assert_called_once_with([(self.test, start, finish, skip.key)])
    
    def test_passes_on_unexpected_success_to_database(self, Database):
        db = Database()
        result, start, finish = self.start_test(db)
        
        result.addUnexpectedSuccess(self.test)
        self.stop_test(result)
        
        db.add_results.assert_called_once_with([(self.test, start, finish, unexpected_success.key)])


@mock.patch('autocheck.db.Database')
class DatabaseTestRunnerVows(TestTestCase):
    
    def test_accepts_database_parameter(self, Database):
        db = Database()
        constructor = partial(TestRunner, database=db, stream=self.stream)
        
        constructor |should_not| throw(TypeError)
    
    def test_accepts_full_suite_parameter(self, Database):
        db = Database()
        constructor = partial(TestRunner, full_suite=False, stream=self.stream)
        
        constructor |should_not| throw(TypeError)
    
    def test_passes_on_run_to_database(self, Database):
        db = Database()
        runner = TestRunner(database=db, stream=self.stream, growler=None)
        
        runner.run(self.test)
        
        db.add_run.assert_called_once_with()
        db.finish_run.assert_called_once_with(True)
    
    def test_passes_on_partial_run_to_database(self, Database):
        db = Database()
        runner = TestRunner(database=db, full_suite=False, stream=self.stream, growler=None)
        
        runner.run(self.test)
        
        db.add_run.assert_called_once_with()
        db.finish_run.assert_called_once_with(False)
    
    def test_accepts_now_function_parameter(self, Database):
        def now_function():
            pass
        constructor = partial(TestRunner, now_function=now_function, stream=self.stream)
        
        constructor |should_not| throw(TypeError)

#.............................................................................
#   test_testrunner.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-12.
#=============================================================================
#   autorunner.py --- Run tests automatically
#=============================================================================
from __future__ import print_function, unicode_literals

import os
import subprocess
import threading

from watchdog.events import RegexMatchingEventHandler
from watchdog.utils import has_attribute, unicode_paths

from .contrib.django import is_django
from .gitignore import GitIgnore


DEFAULT_FILEPATTERN = r'.*\.(py|txt|yaml|sql|html|js|css|feature|xml)$'

class AutocheckEventHandler(RegexMatchingEventHandler):
    
    def __init__(self, dir, args, filepattern=DEFAULT_FILEPATTERN, database=None):
        self._lock = threading.Lock()
        self.child = None
        self.args = args
        if is_django():
            settings = None
            for i, arg in enumerate(args[1:]):
                if arg.startswith('--settings'):
                    if arg.startswith('--settings='):
                        settings = arg.split('=', 1)[0]
                    else:
                        settings = arg[i+1]
                    break
            if not settings and os.path.exists('test_settings.py'):
                settings = 'test_settings'
            if settings:
                os.environ['DJANGO_SETTINGS_MODULE'] = settings
            self.args[0:1] = ['./manage.py', 'test']
        else:
            self.args = args + ['--once']
        for arg in args:
            if arg.startswith('--python='):
                self.args = [arg.split('=', 1)[1]] + self.args
        self.db = database
        self.gitignore = GitIgnore(dir)
        super(AutocheckEventHandler, self).__init__(regexes=[filepattern], ignore_directories=True, case_sensitive=False)
    
    def dispatch(self, event):
        paths = []
        if has_attribute(event, 'dest_path'):
            paths.append(unicode_paths.decode(event.dest_path))
        elif event.src_path:
            paths.append(unicode_paths.decode(event.src_path))
        
        if any(self.gitignore.match(p) for p in paths):
            return
        super(AutocheckEventHandler, self).dispatch(event)
    
    @property
    def child(self):
        with self._lock:
            return self._child
    
    @child.setter
    def child(self, child):
        with self._lock:
            self._child = child
    
    def kill_child(self):
        child = self.child
        if child is not None:
            return True
    
    def run_tests(self):
        print(' '.join(self.args))
        self.child = subprocess.Popen(self.args, close_fds=True)
        self.child.wait()
        returncode = self.child.returncode
        self.child = None
        if self.db is None:
            return False
        try:
            return self.db.should_run_again()
        finally:
            self.db.close()
    
    def on_any_event(self, event):
        while self.run_tests():
            pass

#.............................................................................
#   autorunner.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-13.
#=============================================================================
#   growler.py --- Growl notifier
#=============================================================================
from __future__ import print_function, unicode_literals

from os.path import dirname, join

from gntp.notifier import GrowlNotifier


class Notifier(object):
    
    def __init__(self, app_name='autocheck'):
        self.growl = GrowlNotifier(
            applicationName = app_name,
            notifications = ['New Message'],
            defaultNotifications = ['New Message']
        )
        self.growl.register()

    def notify(self, title, description, kind='pass', sticky=False):
        icon = open(join(dirname(__file__), 'images', kind + '.png'), 'rb').read()
        self.growl.notify(
            noteType = 'New Message',
            title = title,
            description = description,
            icon = icon,
            sticky = sticky,
        )

#.............................................................................
#   growler.py

# Smilies by Jamie Hill:
# http://thelucid.com/2007/07/30/autotest-growl-fail-pass-smilies/



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-20.
#=============================================================================
#   testsuiterunner.py --- Django test suite runner
#=============================================================================
from optparse import make_option

from django.test.simple import DjangoTestSuiteRunner

from .quickstart import QuickstartMixin
from autocheck.db import Database
from autocheck.testrunner import TestRunner, TestProgram


class TestSuiteRunner(QuickstartMixin, DjangoTestSuiteRunner):
    opts = dict(
        failfast = '--failfast',
        catch = '--catch',
        buffer = '--buffer',
        pattern = '-p',
        verbose = '-v',
    )
    discover_opts = dict(
        start = '-s',
        top_level = '-t',
    )
    discover_opts.update(opts)
    option_list = (
        make_option('--no-buffer', action='store_false', dest='buffer', default=True,
            help='Do not buffer stdout and stderr during test runs.'),
        make_option('--no-verbose', action='store_false', dest='verbose', default=True,
            help='Do not run unittests verbose.'),
        make_option('--tags', metavar='TAGEXPRESSION', action='append', default=[],
            help='Filter tests by tag expression.'),
        make_option('-s', metavar='DIRECTORY', dest='start', default='.',
            help='Directory to start discovery (default: ".").'
        ),
        make_option('-p', metavar='PATTERN', dest='pattern',
            help='Pattern to match test files ("test*.py" default).'
        ),
        make_option('-t', metavar='DIRECTORY', dest='top_level', default='.',
            help='Top level directory of project (default: ".").'
        ),
    )
    
    def __init__(self, **options):
        super(TestSuiteRunner, self).__init__(**options)
        self.options = options
    
    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        argv = ['./manage.py unittest']
        discover = len(test_labels) == 0 or len(test_labels) == 1 and test_labels[0] == 'discover'
        opts = self.opts.copy()
        if discover:
            argv.append('discover')
            opts.update(self.discover_opts)
        for key, value in self.options.items():
            if key in opts:
                option = opts[key]
                if value is True:
                    argv.append(option)
                elif value is not False and value is not None:
                    argv.append(option)
                    argv.append(value)
        for tag in self.options['tags']:
            argv.extend(('--tags', tag))
        if not discover:
            argv.extend(test_labels)
        self.setup_test_environment()
        old_config = self.setup_databases()
        result = TestProgram(module=None, argv=argv, testRunner=TestRunner, database=Database())
        self.teardown_databases(old_config)
        self.teardown_test_environment()
        return self.suite_result(suite, result)

#.............................................................................
#   testsuiterunner.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-15.
#=============================================================================
#   test_extra_matchers.py ---
#=============================================================================
import mock
from should_dsl import should, should_not

from .extra_matchers import EachEqual
from autocheck.compat import unittest


class EachEqualMatcherTest(unittest.TestCase):
    
    def test_sees_equality(self):
        matcher = EachEqual()(range(10))
        diff = list(matcher.differ(range(10)))
        diff |should| be_equal_to([])
    
    def test_gets_the_differences(self):
        left = [1, 2, 3, 4]
        right = [3, 2, 1, 4]
        matcher = EachEqual()(right)
        diff = list(matcher.differ(left))
        diff |should| be_equal_to([
            (1, 1, 3),
            (3, 3, 1),
        ])
    
    def test_gets_the_differences_with_different_len(self):
        left = [1, 2, 3]
        right = [3, 2, 1, 4]
        matcher = EachEqual()(right)
        diff = list(matcher.differ(left))
        diff |should| be_equal_to([
            (1, 1, 3),
            (3, 3, 1),
            (4, None, 4),
        ])

#.............................................................................
#   test_extra_matchers.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-30.
#=============================================================================
#   test_filtersuite.py --- FilterSuite vows
#=============================================================================
import os.path
import re
from functools import partial

import mock
import six
import yaml
from should_dsl import should
from six.moves import map

from autocheck.compat import unittest
from autocheck.filtersuite import filter_suite, flatten_suite


class FilterSuiteTestCase(unittest.TestCase):
    __suites = yaml.load(open(os.path.splitext(__file__)[0] + '.yaml'))['suites']
    __cache = dict()
    
    class Test(unittest.TestCase):
        
        def runTest(self):
            pass
        
        @classmethod
        def add_test(cls, name):
            setattr(cls, 'test_' + name, cls.runTest)
    
    def get_suite(self, name):
        try:
            suite = self.__cache[name]
        except KeyError:
            suite = self.__cache[name] = self.add_suite(name, self.__suites[name])
        return suite
    
    def add_suite(self, name, defs):
        if not defs:
            return unittest.TestSuite()
        elif isinstance(defs, list):
            return self.add_tests(name, defs)
        elif isinstance(defs, dict):
            return self.add_suites(name, defs)
        else:
            raise RuntimeError('invalid suite definition: %s' % defs)
    
    def add_suites(self, name, defs):
        return unittest.TestSuite(
            self.add_suite('_'.join((name, key)), value)
            for key, value in sorted(defs.items())
        )
    
    def add_tests(self, name, defs):
        class Test(self.Test): pass
        Test.__name__ = 'Test_' + name
        for test_name in defs:
            Test.add_test(test_name)
        return unittest.defaultTestLoader.loadTestsFromTestCase(Test)


class FilterSuiteTestCaseVows(FilterSuiteTestCase):
    
    def test_returns_TestSuite_instance(self):
        suite = self.get_suite('empty')
        
        suite |should| be_instance_of(unittest.TestSuite)
    
    def test_creates_empty_suite_properly(self):
        suite = self.get_suite('empty')
        
        map(str, suite) |should| each_be_equal_to([])
    
    def test_creates_flat_suite_properly(self):
        suite = self.get_suite('flat')
        
        map(suite_str, suite) |should| each_be_equal_to([
    	    'test_one (Test_flat)',
    	    'test_three (Test_flat)',
    	    'test_two (Test_flat)',
        ])
    
    def test_creates_nested_suite_properly(self):
        suite = self.get_suite('nested')
        
        map(suite_str, suite) |should| each_be_equal_to([
            '<TestSuite tests=['
                '<Test_nested_first testMethod=test_one>, '
                '<Test_nested_first testMethod=test_two>]>',
            '<TestSuite tests=['
                '<Test_nested_second testMethod=test_four>, '
                '<Test_nested_second testMethod=test_three>]>',
        ])
    
    def test_creates_mixed_suite_properly(self):
        suite = self.get_suite('mixed')
        
        map(suite_str, suite) |should| each_be_equal_to([
            '<TestSuite tests=['
                '<Test_mixed_flat testMethod=test_one>, '
                '<Test_mixed_flat testMethod=test_two>]>',
            '<TestSuite tests=['
                '<TestSuite tests=['
                    '<Test_mixed_nested_first testMethod=test_four>, '
                    '<Test_mixed_nested_first testMethod=test_three>]>, '
                '<TestSuite tests=['
                    '<Test_mixed_nested_second testMethod=test_five>, '
                    '<Test_mixed_nested_second testMethod=test_six>]>, '
                    '<TestSuite tests=['
                        '<TestSuite tests=['
                            '<Test_mixed_nested_third_deep testMethod=test_eight>, '
                            '<Test_mixed_nested_third_deep testMethod=test_seven>]>]>]>',
        ])


class FlattenSuiteVows(FilterSuiteTestCase):
    
    def test_returns_TestSuite_instance(self):
        suite = self.get_suite('empty')
        
        flattened = flatten_suite(suite)
        
        flattened |should| be_instance_of(unittest.TestSuite)
    
    def test_handles_empty_suite_properly(self):
        suite = self.get_suite('empty')
        
        flattened = flatten_suite(suite)
        
        next = partial(six.next, iter(flattened))
        next |should| throw(StopIteration)
    
    def test_handles_flat_suite_properly(self):
        suite = self.get_suite('flat')
        
        flattened = flatten_suite(suite)
        
        map(suite_str, flattened) |should| each_be_equal_to([
    	    'test_one (Test_flat)',
    	    'test_three (Test_flat)',
    	    'test_two (Test_flat)',
        ])
    
    def test_flattens_nested_suite_properly(self):
        suite = self.get_suite('nested')
        
        flattened = flatten_suite(suite)
        
        map(suite_str, flattened) |should| each_be_equal_to([
            'test_one (Test_nested_first)',
            'test_two (Test_nested_first)',
            'test_four (Test_nested_second)',
            'test_three (Test_nested_second)',
        ])
    
    def test_flattens_mixed_suite_properly(self):
        suite = self.get_suite('mixed')
        
        flattened = flatten_suite(suite)
        
        map(suite_str, flattened) |should| each_be_equal_to([
            'test_one (Test_mixed_flat)',
            'test_two (Test_mixed_flat)',
            'test_four (Test_mixed_nested_first)',
            'test_three (Test_mixed_nested_first)',
            'test_five (Test_mixed_nested_second)',
            'test_six (Test_mixed_nested_second)',
            'test_eight (Test_mixed_nested_third_deep)',
            'test_seven (Test_mixed_nested_third_deep)',
        ])


@mock.patch('autocheck.db.Database')
class FilterSuiteVows(FilterSuiteTestCase):
    
    def test_returns_TestSuite_instance(self, Database):
        suite = self.get_suite('flat')
        candidates = list(map(str, suite))[:1]
        db = Database()
        db.candidates = mock.Mock('candidates', return_value=set(candidates))
        
        filtered, full_suite = filter_suite(suite, db)
        
        filtered |should| be_instance_of(unittest.TestSuite)
    
    def test_returns_all_tests_when_there_is_no_database(self, Database):
        suite = self.get_suite('flat')
        
        filtered, full_suite = filter_suite(suite, None)
        
        full_suite |should| be(True)
        map(str, filtered) |should| each_be_equal_to(map(str, suite))
    
    def test_returns_all_tests_when_candidates_empty(self, Database):
        suite = self.get_suite('flat')
        candidates = set()
        db = Database()
        db.candidates = mock.Mock('candidates', return_value=set())
        
        filtered, full_suite = filter_suite(suite, db)
        
        full_suite |should| be(True)
        map(str, filtered) |should| each_be_equal_to(map(str, suite))
    
    def test_returns_only_tests_that_are_candidates_if_there_are_candidates(self, Database):
        suite = self.get_suite('flat')
        candidates = ['test_two (%s.Test_flat)' % __name__]
        db = Database()
        db.candidates = mock.Mock('candidates', return_value=set(candidates))
        
        filtered, full_suite = filter_suite(suite, db)
        
        full_suite |should| be(False)
        map(str, filtered) |should| each_be_equal_to(candidates)
    
    def test_returns_all_tests_when_filtered_empty(self, Database):
        suite = self.get_suite('flat')
        candidates = ['unknown']
        db = Database()
        db.candidates = mock.Mock('candidates', return_value=set(candidates))
        
        filtered, full_suite = filter_suite(suite, db)
        
        full_suite |should| be(True)
        map(str, filtered) |should| each_be_equal_to(map(str, suite))


def suite_str(suite):
    return re.sub(r'unittest2?\.suite\.', '', str(suite)).replace(__name__+'.', '')

#.............................................................................
#   test_filtersuite.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-06-01.
#=============================================================================
#   test_tags.py --- Tag support vows
#=============================================================================
from should_dsl import should
from six.moves import map

from autocheck.compat import unittest
from autocheck.tagexpression import TagExpression
from autocheck.tags import tag, get_tags


class TagDecoratorVows(unittest.TestCase):
    
    def test_markes_test_item_by_test_case_decorator(self):
        @tag('include')
        class Test(unittest.TestCase):
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['include']))
    
    def test_markes_test_item_by_test_method_decorator(self):
        class Test(unittest.TestCase):
            @tag('include')
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['include']))
    
    def test_combines_tags_from_method_and_class(self):
        @tag('class')
        class Test(unittest.TestCase):
            @tag('method')
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['class', 'method']))
    
    def test_combines_tags_from_class_and_superclass(self):
        @tag('super')
        class Super(unittest.TestCase):
            pass
        @tag('class')
        class Test(Super):
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['class', 'super']))
    
    def test_combines_tags_from_method_and_superclass(self):
        @tag('super')
        class Super(unittest.TestCase):
            pass
        class Test(Super):
            @tag('method')
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['method', 'super']))

    def test_combines_tags_from_class_and_mixin(self):
        @tag('mixin')
        class Mixin(object):
            pass
        @tag('class')
        class Test(Mixin, unittest.TestCase):
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['class', 'mixin']))

    def test_combines_tags_from_class_superclass_and_mixin(self):
        @tag('mixin')
        class Mixin(object):
            pass
        @tag('super')
        class Super(unittest.TestCase):
            pass
        @tag('class')
        class Test(Mixin, Super):
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['class', 'mixin', 'super']))

    def test_combines_tags_from_method_class_superclass_and_mixin(self):
        @tag('mixin')
        class Mixin(object):
            pass
        @tag('super')
        class Super(unittest.TestCase):
            pass
        @tag('class')
        class Test(Mixin, Super):
            @tag('method')
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['class', 'method', 'mixin', 'super']))

    def test_does_not_mix_tags_from_different_classes_with_same_mixin(self):
        @tag('mixin')
        class Mixin(object):
            pass
        @tag('other')
        class Other(Mixin, unittest.TestCase):
            pass
        @tag('class')
        class Test(Mixin, unittest.TestCase):
            @tag('method')
            def test_method(self):
                pass
        test_item = Test('test_method')
        
        get_tags(test_item) |should| be_equal_to(set(['class', 'method', 'mixin']))


class TagFilterVows(unittest.TestCase):
    
    def test_includes_by_test_case_decorator(self):
        @tag('include')
        class Test(unittest.TestCase):
            def test_method(self):
                pass
        suite = unittest.TestSuite([Test('test_method')])
        tags = TagExpression('@include')
        
        filtered = tags.filter_suite(suite)
        
        map(str, filtered) |should| each_be_equal_to([
            'test_method (vows.test_tags.Test)',
        ])
    
    def test_excludes_by_test_case_decorator(self):
        @tag('include')
        class Test(unittest.TestCase):
            def test_method(self):
                pass
        suite = unittest.TestSuite([Test('test_method')])
        tags = TagExpression('~@include')
        
        filtered = tags.filter_suite(suite)
        
        map(str, filtered) |should| each_be_equal_to([])
    
    def test_includes_by_test_method_decorator(self):
        class Test(unittest.TestCase):
            @tag('include')
            def test_method(self):
                pass
        suite = unittest.TestSuite([Test('test_method')])
        tags = TagExpression('@include')
        
        filtered = tags.filter_suite(suite)
        
        map(str, filtered) |should| each_be_equal_to([
            'test_method (vows.test_tags.Test)',
        ])
    
    def test_excludes_by_test_method_decorator(self):
        class Test(unittest.TestCase):
            @tag('include')
            def test_method(self):
                pass
        suite = unittest.TestSuite([Test('test_method')])
        tags = TagExpression('~@include')
        
        filtered = tags.filter_suite(suite)
        
        map(str, filtered) |should| each_be_equal_to([])


#.............................................................................
#   test_tags.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-29.
#=============================================================================
#   colorizer.py --- Colours
#=============================================================================
from __future__ import print_function, unicode_literals

import os.path

import six
import termstyle
import yaml


class ColourDecorator(object):
    """Used to decorate file-like objects' 'write' method to accept colours"""
    def __init__(self, stream):
        self.stream = stream
    
    def __getattr__(self, attr):
        if attr in ('stream', '__getstate__'):
            raise AttributeError(attr)
        return getattr(self.stream, attr)
    
    def write(self, arg=None, colour=None):
        if arg:
            if colour is not None:
                arg = colour(arg)
        self.stream.write(arg)


class ColourWritelnDecorator(ColourDecorator):
    """Used to decorate file-like objects with a handy 'writeln' method"""
    def __init__(self, stream):
        self.stream = stream
    
    def __getattr__(self, attr):
        if attr in ('stream', '__getstate__'):
            raise AttributeError(attr)
        return getattr(self.stream, attr)
    
    def writeln(self, arg=None, colour=None):
        if arg:
            self.write(arg, colour=colour)
        self.write('\n') # text-mode streams translate to \r\n if needed


class ColourScheme(object):
    
    def __init__(self, scheme):
        if isinstance(scheme, six.string_types):
            filename = os.path.join(os.path.dirname(__file__), 'colourschemes', scheme + '.yaml')
            with open(filename) as scheme_file:
                self.scheme = yaml.load(scheme_file)
        else:
            self.scheme = scheme
    
    def __getattr__(self, key):
        if self.scheme is not None:
            colour = self.scheme.get(key)
            if isinstance(colour, six.string_types):
                return getattr(termstyle, colour)
            elif colour is None:
                return lambda x: x
            else:
                return compose(getattr(termstyle, c) for c in colour)

def compose(iterable):
    def compose(arg):
        for f in iterable:
            arg = f(arg)
        return arg
    return compose


#.............................................................................
#   colorizer.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-29.
#=============================================================================
#   extra_matchers.py --- Should-DSL extra matchers
#=============================================================================
from should_dsl import matcher
from six.moves import zip_longest


@matcher
class EachEqual(object):
    name = 'each_be_equal_to'
    
    def __call__(self, expected):
        self._expected = expected
        return self
    
    def differ(self, given):
        for n, (left, right) in enumerate(zip_longest(given, self._expected)):
            if left != right:
                yield n + 1, left, right
    
    def match(self, given):
        diff = list(self.differ(given))
        self.diff = '\n\t'.join('%d: %r is not equal to %r' % item for item in diff)
        return not diff
    
    def message_for_failed_should(self):
        return 'sequences differ\n\t' + self.diff


@matcher
def be_equal_to():
    return (lambda x, y: x == y, '%r is %sequal to %r')

@matcher
def be_in():
    return (lambda item, container: item in container, '%r is %sinto %r')

#.............................................................................
#   extra_matchers.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-12-11.
#=============================================================================
#   setup.py --- Install script
#=============================================================================
import sys
from distutils import spawn
from setuptools import setup


install_requires=['PyYAML', 'watchdog', 'python-termstyle', 'six']

if sys.version_info[:2] == (2, 6):
    install_requires.append('unittest2')

if spawn.find_executable('pandoc'):
    extra_setup = dict(
        setup_requires=['setuptools-markdown'],
        long_description_markdown_filename='README.md',        
    )
else:
    extra_setup = {}

setup(
    name='autocheck',
    version='0.2.5',
    description='Improved unittest test runner',
    author='Hans-Thomas Mueller',
    author_email='htmue@mac.com',
    url='https://github.com/htmue/python-autocheck',
    packages=['autocheck',
        'autocheck.contrib',
        'autocheck.contrib.django',
    ],
    package_data=dict(autocheck=['*.yaml', 'colourschemes/*.yaml', 'images/*.png']),
    scripts=['bin/autocheck'],
    install_requires=install_requires,
    **extra_setup
)

#.............................................................................
#   setup.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2015-04-19.
#=============================================================================
#   discoveryrunner.py --- Django test suite runner
#=============================================================================
from django.test.runner import DiscoverRunner, reorder_suite

from .quickstart import QuickstartMixin
from autocheck.db import Database
from autocheck.tagexpression import TagExpression
from autocheck.testrunner import TestRunner


class TestSuiteRunner(QuickstartMixin, DiscoverRunner):
    test_runner = TestRunner
    
    def __init__(self, buffer=True, tags=[], **kwargs):
        super(TestSuiteRunner, self).__init__(**kwargs)
        self.buffer = buffer
        verbose = kwargs.pop('verbose', None)
        if verbose and self.verbosity < 2:
            self.verbosity = 2
        self.run_tags = None
        for tag in tags:
            if self.run_tags is None:
                self.run_tags = TagExpression()
            self.run_tags.parse_and_add(tag)

    @classmethod
    def add_arguments(cls, parser):
        super(TestSuiteRunner, cls).add_arguments(parser)
        parser.add_argument('--no-buffer', action='store_false', dest='buffer', default=True,
            help='Do not buffer stdout and stderr during test runs.')
        parser.add_argument('--no-verbose', action='store_false', dest='verbose', default=True,
            help='Do not run unittests verbose.')
        parser.add_argument('--tags', metavar='TAGEXPRESSION', action='append', default=[],
            help='Filter tests by tag expression.')

    def build_suite(self, test_labels=None, extra_tests=None, **kwargs):
        suite = super(TestSuiteRunner, self).build_suite(test_labels, extra_tests, **kwargs)
        if self.run_tags is not None:
            suite = self.run_tags.filter_suite(suite)
        return reorder_suite(suite, self.reorder_by, self.reverse)

    def run_suite(self, suite, **kwargs):
        resultclass = self.get_resultclass()
        return self.test_runner(
            buffer=self.buffer,
            verbosity=self.verbosity,
            failfast=self.failfast,
            resultclass=resultclass,
            database=Database(),
        ).run(suite)

#.............................................................................
#   discoveryrunner.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-12.
#=============================================================================
#   main.py --- Run tests automatically
#=============================================================================
from __future__ import print_function, unicode_literals

import os
import signal
import sys
import time

from watchdog.observers.fsevents import FSEventsObserver as Observer

from .autorunner import AutocheckEventHandler
from .db import Database
from .testrunner import TestRunner, TestProgram


class Unbuffered:
    
    def __init__(self, stream):
        self.stream = stream
    
    def write(self, data):
        self.stream.write(data)
        self.stream.flush()
    
    def __getattr__(self, attr):
        return getattr(self.stream, attr)

def handle_term():
    def sighandler(signum, frame):
        os.kill(os.getpid(), signal.SIGINT)
    signal.signal(signal.SIGTERM, sighandler)

def single(args):
    sys.stdout=Unbuffered(sys.stdout)
    TestProgram(module=None, testRunner=TestRunner, argv=args, database=Database())

def autocheck(args):
    handle_term()
    event_handler = AutocheckEventHandler('.', args, database=Database())
    try:
        event_handler.run_tests()
    except KeyboardInterrupt:
        pass
    observer = Observer()
    observer.schedule(event_handler, '.', recursive=True)
    observer.start()
    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            if not event_handler.kill_child():
                print('Got signal, exiting.', file=sys.stderr)
                observer.stop()
                break
    observer.join()

def main(args=sys.argv):
    import resource
    try:
        resource.setrlimit(resource.RLIMIT_NOFILE, (4096, -1))
    except Exception as e:
        print(repr(e))
    if '--once' in args:
        if args[1] == '-m' and args[2] in ('autocheck', 'autocheck.main'):
            args[1:3] = []
        args.remove('--once')
        single(args)
    elif '--stats-flat' in args:
        database = Database()
        for test in database.stats():
            print('{time:f} {runs: >8}\t{suite}.{test}'.format(**test))
    elif '--stats' in args:
        database = Database()
        total_time = 0.0
        for suite in database.stats_grouped():
            print('{time:f} {suite}'.format(**suite))
            total_time += suite['time']
            for test in suite['tests']:
                print('\t{time:f} {runs: >8}\t{test}'.format(**test))
        print('total: {:f}'.format(total_time))
    else:
        autocheck(args)

if __name__ == '__main__':
    if sys.argv[0].endswith('main.py'):
        sys.argv[0:1] = [os.path.abspath(sys.executable), '-m', 'autocheck.main']
    
    main()

#.............................................................................
#   main.py



# -*- coding:utf-8 -*-
# Created by Hans-Thomas on 2011-05-28.
#=============================================================================
#   test_db.py --- Tests database vows
#=============================================================================
import datetime
from functools import partial

from should_dsl import should, should_not

from autocheck.compat import unittest
from autocheck.db import Database, timedelta_to_float, RunDoesNotExist, ResultDoesNotExist
from autocheck.status import ok, error, fail, skip


class DatabaseVows(unittest.TestCase):
    class Test(unittest.TestCase):
        def test_name(self): pass
        def test_name_0(self): pass
        def test_name_1(self): pass
        def test_name_2(self): pass
        def test_name_3(self): pass
        def test_name_0_ok(self): pass
        def test_name_1_FAIL(self): pass
        def test_name_2_ERROR(self): pass
        def test_name_3_ok(self): pass
        def test_name_4_skipped(self): pass
    
    def setUp(self):
        self.db = Database(path=':memory:')
    
    def test_can_connect_to_database(self):
        self.db.connect()
        
        self.db |should| be_connected
    
    def test_can_close_database(self):
        self.db.connect()
        
        self.db.close()
        
        self.db |should_not| be_connected
    
    def test_can_add_test_run(self):
        self.db.add_run()
        
        self.db.current_run_id |should| be(1)
        self.db.total_runs |should| be(1)
    
    def test_can_count_test_runs(self):
        for i in range(5):
            self.db.add_run()
            
            self.db.current_run_id |should| be(i + 1)
            self.db.total_runs |should| be(i + 1)
    
    def test_can_load_test_run_object(self):
        before = datetime.datetime.utcnow()
        self.db.add_run()
        after = datetime.datetime.utcnow()
        
        run = self.db.get_run()
        
        run['id'] |should| be(1)
        run['started'] |should| be_greater_than(before)
        run['started'] |should| be_less_than(after)
    
    def test_raises_RunDoesNotExist_when_no_run_exists(self):
        self.db.get_run |should| throw(RunDoesNotExist)
    
    def test_raises_RunDoesNotExist_when_run_does_not_exist(self):
        get_run = partial(self.db.get_run, 1)
        
        get_run |should| throw(RunDoesNotExist)
    
    def add_result(self, test, status=ok.key, started=None, finished=None, run=None):
        if run is None:
            self.db.add_run()
            run = self.db.get_run()
        if finished is None:
            finished = datetime.datetime.utcnow()
        if started is None:
            started = run['started']
        self.db.add_result(test, started, finished, status)
    
    def test_can_add_result(self):
        test = self.Test('test_name')
        
        self.add_result(test)
        
        self.db.total_runs_by_test_name(str(test)) |should| be(1)
    
    def test_raises_ResultDoesNotExist_when_test_does_not_exist(self):
        get_result = partial(self.db.get_last_result, 'hi')
        
        get_result |should| throw(ResultDoesNotExist)
    
    def test_knows_number_of_test_runs(self):
        test = self.Test('test_name')
        self.add_result(test)
        self.add_result(test)
        
        self.db.total_runs_by_test_name(str(test)) |should| be(2)
    
    def test_knows_average_test_runtime(self):
        test = self.Test('test_name')
        now = datetime.datetime.utcnow()
        started = now
        finished = started + datetime.timedelta(seconds=1)
        self.add_result(test, started=started, finished=finished)
        started = finished
        finished = started + datetime.timedelta(seconds=2)
        self.add_result(test, started=started, finished=finished)
        
        test = self.db.get_test(str(test))
        
        average_time = timedelta_to_float(test['average_time'])
        average_time |should| close_to(1.5, delta=.1)
    
    def test_knows_last_run_of_test_run(self):
        test = self.Test('test_name')
        self.add_result(test)
        self.add_result(test)
        
        result = self.db.get_last_result(str(test))
        
        result['run_id'] |should| be_equal_to(self.db.current_run_id)
    
    def test_can_finish_test_run(self):
        test = self.Test('test_name')
        self.add_result(test)
        
        run = self.db.finish_run(True)
        
        run['testsRun'] |should| be(1)
        run['wasSuccessful'] |should| be(True)
        run['errors'] |should| be(0)
        run['failures'] |should| be(0)
        run['skipped'] |should| be(0)
        run['expectedFailures'] |should| be(0)
        run['unexpectedSuccesses'] |should| be(0)
    
    def test_knows_id_of_last_successful_run(self):
        test = self.Test('test_name')
        self.add_result(test, status=ok.key)
        self.db.finish_run(False)
        self.add_result(test, status=fail.key)
        self.db.finish_run(False)
        
        self.db.get_last_successful_run_id() |should| be(1)
    
    def test_handles_last_successful_run_id_edge_case_empty(self):
        self.db.get_last_successful_run_id() |should| be(None)
    
    def test_handles_last_successful_run_id_edge_case_only_failures(self):
        test = self.Test('test_name')
        self.add_result(test, status=fail.key)
        self.db.finish_run(True)
        
        self.db.get_last_successful_run_id() |should| be(None)
    
    def test_cleans_history_when_finishing_run(self):
        test = self.Test('test_name')
        self.add_result(test)
        self.db.finish_run(False)
        self.add_result(test)
        self.db.finish_run(True)
        self.add_result(test)
        self.db.finish_run(False)
        
        self.db.total_runs |should| be(2)
    
    def test_does_not_clean_history_after_unsuccessful_run(self):
        test = self.Test('test_name')
        self.add_result(test, status=ok.key)
        self.db.finish_run(True)
        self.add_result(test, status=fail.key)
        self.db.finish_run(True)
        
        self.db.total_runs |should| be(2)
    
    def test_does_not_clean_history_after_partial_run(self):
        test = self.Test('test_name')
        self.add_result(test, status=ok.key)
        self.db.finish_run(True)
        self.add_result(test, status=ok.key)
        self.db.finish_run(False)
        
        self.db.total_runs |should| be(2)
    
    def test_keeps_track_of_full_test_runs(self):
        test = self.Test('test_name')
        self.add_result(test)
        run = self.db.finish_run(True)
        
        run['full'] |should| be(True)
    
    def test_keeps_track_of_partial_test_runs(self):
        test = self.Test('test_name')
        self.add_result(test)
        run = self.db.finish_run(False)
        
        run['full'] |should| be(False)
    
    def test_knows_id_of_last_run(self):
        for d in range(3):
            test = self.Test('test_name_%d' % d)
            self.add_result(test)
            run = self.db.finish_run(True)
        
        self.db.get_last_run_id() |should| be_equal_to(run['id'])
    
    def test_knows_id_of_last_successful_full_run(self):
        name = 'test_name_%d'
        self.add_result(self.Test(name % 1), status=ok.key)
        run_1 = self.db.finish_run(True)
        self.add_result(self.Test(name % 2), status=ok.key)
        run_2 = self.db.finish_run(False)
        self.add_result(self.Test(name % 3), status=fail.key)
        run_3 = self.db.finish_run(True)
        
        self.db.get_last_successful_full_run_id() |should| be_equal_to(run_1['id'])
    
    def test_can_collect_non_successful_tests_after_certain_run(self):
        name = 'test_name_%d_%s'
        run_ids = []
        for i, status in enumerate((ok, fail, error, ok, skip)):
            self.add_result(self.Test(name % (i, status.name)), status=status.key)
            run_ids.append(self.db.finish_run(False)['id'])
        
        results = list(self.db.collect_results_after(run_ids[0]))
        results |should| each_be_equal_to([
            'test_name_1_FAIL (vows.test_db.Test)',
            'test_name_2_ERROR (vows.test_db.Test)',
            'test_name_4_skipped (vows.test_db.Test)',
        ])
        
        results = list(self.db.collect_results_after(run_ids[2]))
        results |should| each_be_equal_to([
            'test_name_4_skipped (vows.test_db.Test)',
        ])
        
        results = list(self.db.collect_results_after(run_ids[4]))
        results |should| be_empty
    
    def test_can_collect_non_successful_tests_after_certain_run_that_have_been_successful_later(self):
        test = self.Test('test_name')
        run_ids = []
        for status in (ok, fail, error, ok, skip):
            self.add_result(test, status=status.key)
            run_ids.append(self.db.finish_run(False)['id'])
        
        results = list(self.db.collect_results_after(run_ids[0]))
        results |should| each_be_equal_to([
            'test_name (vows.test_db.Test)',
        ])
        
        results = list(self.db.collect_results_after(run_ids[2]))
        results |should| each_be_equal_to([
            'test_name (vows.test_db.Test)',
        ])
        
        results = list(self.db.collect_results_after(run_ids[4]))
        results |should| be_empty
    
    def test_can_collect_new_tests(self):
        tests = [self.Test('test_name_%d' % i) for i in range(4)]
        suite = unittest.TestSuite(tests[:2])
        
        results = list(self.db.collect_changes(suite))
        results |should| each_be_equal_to([
            'test_name_0 (vows.test_db.Test)',
            'test_name_1 (vows.test_db.Test)',
        ])
        
        self.add_result(tests[0])
        self.add_result(tests[1])
        suite = unittest.TestSuite(tests)
        
        results = list(self.db.collect_changes(suite))
        results |should| each_be_equal_to([
            'test_name_2 (vows.test_db.Test)',
            'test_name_3 (vows.test_db.Test)',
        ])
    
    def test_can_collect_tests_that_have_changed(self):
        tests = [self.Test('test_name_%d' % i) for i in range(4)]
        for test in tests:
            self.add_result(test)
        
        self.Test.test_name_2 = lambda self: None
        tests = [self.Test('test_name_%d' % i) for i in range(4)]
        suite = unittest.TestSuite(tests)
        
        results = list(self.db.collect_changes(suite))
        results |should| each_be_equal_to([
            'test_name_2 (vows.test_db.Test)',
        ])

#.............................................................................
#   test_db.py
